"use strict";
var _a;
exports.__esModule = true;
exports.EMPTY_ITER = exports.EMPTY_ITER_NEXT = exports.ITER_DONE_VALUE = exports.double = exports.cycle = exports.once = exports.empty = exports.flatMap = exports.map = exports.zip = exports.append = exports.chain = void 0;
/**
 * Chains two iterables.
 * @param firstIterable
 * @param secondIterable
 * @returns {Iterable<V>}
 */
function chain(firstIterable, secondIterable) {
    var _a;
    if (firstIterable === exports.EMPTY_ITER) {
        return secondIterable;
    }
    else if (secondIterable === exports.EMPTY_ITER) {
        return firstIterable;
    }
    return _a = {},
        _a[Symbol.iterator] = function () {
            var iter = firstIterable[Symbol.iterator]();
            var firstDone = false;
            return {
                next: function () {
                    for (;;) {
                        var _a = iter.next(), value = _a.value, done = _a.done;
                        if (done) {
                            if (!firstDone) {
                                iter = secondIterable[Symbol.iterator]();
                                firstDone = true;
                            }
                            else {
                                return exports.ITER_DONE_VALUE;
                            }
                        }
                        else {
                            return { value: value, done: false };
                        }
                    }
                }
            };
        },
        _a;
}
exports.chain = chain;
/**
 * Creates an iterable that then iterates over supplied iterable and then emits one item.
 * @param value
 * @param iterable
 * @returns {Iterable<V>}
 */
function append(iterable, value) {
    var _a;
    return _a = {},
        _a[Symbol.iterator] = function () {
            var state = 0;
            var iter = iterable[Symbol.iterator]();
            return {
                next: function () {
                    if (state === 0) {
                        var item = iter.next();
                        if (item.done) {
                            state = 1;
                        }
                        else {
                            return item;
                        }
                    }
                    if (state === 1) {
                        state = 2;
                        return { value: value, done: false };
                    }
                    return exports.ITER_DONE_VALUE;
                }
            };
        },
        _a;
}
exports.append = append;
/**
 * Zips two iterables together.
 * @param leftIterable
 * @param rightIterable
 * @returns {Iterable<{ left: L; right: R }>}
 */
function zip(leftIterable, rightIterable) {
    var _a;
    return _a = {},
        _a[Symbol.iterator] = function () {
            var leftIter = leftIterable[Symbol.iterator]();
            var rightIter = rightIterable[Symbol.iterator]();
            return {
                next: function () {
                    for (;;) {
                        var left = leftIter.next();
                        var right = rightIter.next();
                        if (left.done || right.done) {
                            return exports.ITER_DONE_VALUE;
                        }
                        else {
                            return {
                                value: { left: left.value, right: right.value },
                                done: false
                            };
                        }
                    }
                }
            };
        },
        _a;
}
exports.zip = zip;
/**
 * Maps supplied iterable using given function.
 * @param fn
 * @param iterable
 * @returns {Iterable<R>}
 */
function map(fn, iterable) {
    var _a;
    return _a = {},
        _a[Symbol.iterator] = function () {
            var iter = iterable[Symbol.iterator]();
            return {
                next: function () {
                    var _a = iter.next(), value = _a.value, done = _a.done;
                    return done ? exports.ITER_DONE_VALUE : { value: fn(value), done: done };
                }
            };
        },
        _a;
}
exports.map = map;
/**
 * Flattens iterable mapped with supplied function.
 * @param iterable
 * @param map
 * @returns {Iterable<R>}
 */
function flatMap(map, iterable) {
    var _a;
    return _a = {},
        _a[Symbol.iterator] = function () {
            var rootIter = iterable[Symbol.iterator]();
            var lastIter = null;
            return {
                next: function () {
                    for (;;) {
                        if (lastIter == null) {
                            var _a = rootIter.next(), nextIter = _a.value, done = _a.done;
                            if (!done) {
                                lastIter = map(nextIter)[Symbol.iterator]();
                            }
                            else {
                                return exports.ITER_DONE_VALUE;
                            }
                        }
                        if (lastIter) {
                            var item = lastIter.next();
                            if (item.done) {
                                lastIter = null;
                            }
                            else {
                                return { value: item.value, done: false };
                            }
                        }
                    }
                }
            };
        },
        _a;
}
exports.flatMap = flatMap;
/**
 * Creates an iterable which emits no items.
 * @returns {Iterable<V>}
 */
function empty() {
    return exports.EMPTY_ITER;
}
exports.empty = empty;
/**
 * Creates an iterable which emits one item.
 * @param value
 * @returns {Iterable<V>}
 */
function once(value) {
    var _a;
    return _a = {},
        _a[Symbol.iterator] = function () {
            var taken = false;
            return {
                next: function () {
                    if (!taken) {
                        taken = true;
                        return { value: value, done: false };
                    }
                    else {
                        return exports.ITER_DONE_VALUE;
                    }
                }
            };
        },
        _a;
}
exports.once = once;
/**
 * Creates an iterable that endlessly repeats supplied iterable.
 * @param iter
 * @returns {Iterable<V>}
 */
function cycle(iter) {
    var _a;
    return _a = {},
        _a[Symbol.iterator] = function () {
            var lastIter = iter[Symbol.iterator]();
            return {
                next: function () {
                    for (var i = 0; i++ < 2;) {
                        var item = lastIter.next();
                        if (item.done) {
                            lastIter = iter[Symbol.iterator]();
                        }
                        else {
                            return item;
                        }
                    }
                    return exports.ITER_DONE_VALUE;
                }
            };
        },
        _a;
}
exports.cycle = cycle;
/**
 * Creates an iterable which emits two items.
 * @param first
 * @param second
 * @returns {Iterable<V>}
 */
function double(first, second) {
    var _a;
    return _a = {},
        _a[Symbol.iterator] = function () {
            var state = 0;
            return {
                next: function () {
                    if (state === 0) {
                        state = 1;
                        return { value: first, done: false };
                    }
                    if (state === 1) {
                        state = 2;
                        return { value: second, done: false };
                    }
                    return exports.ITER_DONE_VALUE;
                }
            };
        },
        _a;
}
exports.double = double;
exports.ITER_DONE_VALUE = { value: undefined, done: true };
exports.EMPTY_ITER_NEXT = {
    next: function () {
        return exports.ITER_DONE_VALUE;
    }
};
/**
 * An iterator producing no values.
 */
exports.EMPTY_ITER = (_a = {},
    _a[Symbol.iterator] = function () {
        return exports.EMPTY_ITER_NEXT;
    },
    _a);
//# sourceMappingURL=iterators.js.map