"use strict";
exports.__esModule = true;
exports.Root = void 0;
var value_1 = require("../value");
var iterators_1 = require("../iterators");
var utils_1 = require("../utils");
/**
 * Contains either a value or a pointer.
 */
var ResultOrPointer = /** @class */ (function () {
    function ResultOrPointer() {
        this.result = value_1.NO_VALUE;
        this.ptr = void 0;
    }
    /**
     * Produces a result.
     * @param result
     */
    ResultOrPointer.fromResult = function (result) {
        var that = new ResultOrPointer();
        that.result = result;
        return that;
    };
    /**
     * Produces a pointer.
     * @param ptr
     */
    ResultOrPointer.fromPointer = function (ptr) {
        var that = new ResultOrPointer();
        that.ptr = ptr;
        return that;
    };
    return ResultOrPointer;
}());
/**
 * Weakly or strongly linked value.
 */
var LinkedValue = /** @class */ (function () {
    function LinkedValue(value, weak) {
        this.value = value;
        this.weak = weak;
    }
    return LinkedValue;
}());
/**
 * Stores last path to value including storage nodes.
 * Used to reduce the number of access operations in case the prefix has some elements in common.
 */
var Last = /** @class */ (function () {
    function Last(length, root) {
        this.storages = Array(length);
        this.storages[0] = new LinkedValue(root, false);
        this.path = utils_1.EMPTY_ARRAY;
        this.value = value_1.NO_VALUE;
    }
    /**
     * Associates storage with the provided index.
     * @param idx
     * @param storage
     */
    Last.prototype.setStorage = function (idx, storage) {
        this.storages[idx] = storage;
    };
    /**
     * Retrieves storage using supplied index.
     * @param idx
     */
    Last.prototype.storage = function (idx) {
        return this.storages[idx];
    };
    /**
     * Returns a value in case the path is the same as the previous one.
     * Otherwise returns an index of the greatest common node for provided and stored paths.
     * @param path
     */
    Last.prototype.get = function (path) {
        var length = this.path.length;
        if (length === path.length) {
            var idx = -1;
            for (var i = 0; i < length && (0, value_1.equals)(this.path[i], path[i]); i++) {
                idx = i;
            }
            if (idx === length - 1) {
                return ResultOrPointer.fromResult(this.value);
            }
            return ResultOrPointer.fromPointer(idx === -1 ? 0 : idx);
        }
        else {
            return ResultOrPointer.fromPointer(0);
        }
    };
    /**
     * Updates path and stored value.
     * @param path
     * @param value
     */
    Last.prototype.update = function (path, value) {
        this.path = path;
        this.value = value;
    };
    /**
     * Resets all underlying values.
     */
    Last.prototype.reset = function () {
        var length = this.storages.length;
        this.storages.length = 1;
        this.storages.length = length;
        this.value = value_1.NO_VALUE;
        this.path = utils_1.EMPTY_ARRAY;
    };
    return Last;
}());
var Root = /** @class */ (function () {
    function Root(params, ctx) {
        this.length = params.length;
        this.ctx = ctx;
        this.root = (params.length > 1 ? ctx.createStorage() : ctx.createLeafStorage());
        this.last = new Last(params.length, this.root);
        this.rootPath = new utils_1.ParentPath(this.ctx.rootStorageStrategy, value_1.NO_VALUE);
        this.leafPath = new utils_1.ParentPath(this.ctx.rootLeafStrategy, value_1.NO_VALUE);
        this.getOrInsertWith = params.checkLast
            ? this.checkLastThenGetOrInsertWith
            : this.getOrInsertWith;
    }
    /**
     * Retrieves cached value if already calculated, otherwise calls the supplied function with provided arguments and
     * saves the result.
     * @param path
     * @param calculate
     */
    Root.prototype.getOrInsertWith = function (path, calculate) {
        var length = this.length;
        if (path.length !== length)
            throw new Error("Invalid path length");
        var result = this.extractOrSetPath(path, calculate, 0);
        this.last.reset();
        return result;
    };
    /**
     * - First, checks last arguments to be the same as the provided, and if so, returns cached value.
     * - Second, attempts to retrieve the cached value, otherwise calls the supplied function with provided arguments and
     * saves the result.
     * @param path
     * @param calculate
     */
    Root.prototype.checkLastThenGetOrInsertWith = function (path, calculate) {
        var _a;
        var length = this.length;
        if (path.length !== length)
            throw new Error("Invalid path length");
        var result = value_1.NO_VALUE, ptr = 0;
        (_a = this.last.get(path), result = _a.result, ptr = _a.ptr);
        if (result !== value_1.NO_VALUE) {
            this.readCache();
        }
        else {
            result = this.extractOrSetPath(path, calculate, ptr);
        }
        return result;
    };
    Root.prototype.extractOrSetPath = function (path, calculate, from) {
        if (from === void 0) { from = 0; }
        var _a = this.extractPath(path, from), result = _a.result, ptr = _a.ptr;
        if (result !== value_1.NO_VALUE) {
            this.last.update(path, result);
            this.readCache();
            return result;
        }
        result = calculate.apply(null, path);
        this.setPath(path, result, ptr);
        this.last.update(path, result);
        this.writeCache();
        return result;
    };
    Root.prototype.extractPath = function (path, from) {
        if (from === void 0) { from = 0; }
        var length = this.length;
        var node = this.last.storage(from);
        var idx = from, lessThanPath = true, res = node.value, isLastWeak = node.weak;
        do {
            this.last.setStorage(idx, new LinkedValue(res, isLastWeak));
            isLastWeak = res.isWeak(path[idx]);
            res = res.get(path[idx]);
            lessThanPath = ++idx < length;
        } while (lessThanPath && res !== value_1.NO_VALUE);
        if (lessThanPath || res === value_1.NO_VALUE) {
            return ResultOrPointer.fromPointer(Math.max(0, idx - 1));
        }
        return ResultOrPointer.fromResult(res);
    };
    Root.prototype.setPath = function (path, value, from) {
        if (from === void 0) { from = 0; }
        var length = this.length;
        var idx = from, cache = this.last.storage(from).value;
        while (idx < length - 1) {
            var current = path[idx];
            var next = cache.get(current);
            var isWeak = cache.isWeak(current);
            if (next === value_1.NO_VALUE) {
                var rootPath = isWeak
                    ? (0, iterators_1.once)(this.rootPath)
                    : (0, iterators_1.double)(this.rootPath, new utils_1.ParentPath(cache, current));
                next =
                    idx < length - 2
                        ? this.ctx.createStorage(rootPath)
                        : this.ctx.createLeafStorage((0, iterators_1.append)(rootPath, this.leafPath));
                cache.set(current, next);
            }
            cache = next;
            this.last.setStorage(++idx, new LinkedValue(cache, isWeak));
        }
        cache.set(path[length - 1], value);
    };
    Root.prototype.readCache = function () {
        for (var i = this.length; i-- > 0;) {
            var node = this.last.storage(i);
            if (!node.weak)
                this.ctx.rootStorageStrategy.read(node.value);
        }
        var leafNode = this.last.storage(this.length - 1);
        if (!leafNode.weak) {
            this.ctx.rootLeafStrategy.read(leafNode.value);
        }
    };
    Root.prototype.writeCache = function () {
        for (var i = this.length; i-- > 0;) {
            var node = this.last.storage(i);
            if (!node.weak)
                this.ctx.rootStorageStrategy.write(node.value);
        }
        var leafNode = this.last.storage(this.length - 1);
        if (!leafNode.weak) {
            this.ctx.rootLeafStrategy.write(leafNode.value);
        }
    };
    return Root;
}());
exports.Root = Root;
//# sourceMappingURL=Root.js.map